'''Tests for django-autoconfig.'''
# pylint: disable=C0103
# pylint: disable=R0904

from django_autoconfig import autoconfig

import copy
from django.core.exceptions import ImproperlyConfigured
import django.core.urlresolvers
from django.core.urlresolvers import resolve
from django import test

class ConfigureSettingsTestCase(test.TestCase):
    '''Test the configure_settings method.'''

    BASE_SETTINGS = {
        'LIST_SETTING': [1, 2],
        'BOOLEAN_SETTING': True,
        'DICT_SETTING': {
            'key1': 'value1',
        },
    }

    def setUp(self):
        self.settings_dict = copy.deepcopy(self.BASE_SETTINGS)

    def test_list_merging(self):
        '''
        Test that list settings are merged correctly
        '''

        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_list']
        autoconfig.configure_settings(self.settings_dict)
        self.assertEqual(self.settings_dict['LIST_SETTING'], [1, 2, 3])

    def test_new_setting(self):
        '''
        A new setting (i.e. not in the DJANGO_SETTINGS_MODULE)
        should just end up as the new value.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_new_setting']
        autoconfig.configure_settings(self.settings_dict)
        self.assertEqual(self.settings_dict['NEW_LIST_SETTING'], [1, 2, 3])

    def test_list_setting_from_defaults(self):
        '''
        A list setting that exists in the django.conf.settings.global_settings
        should merge with the default, not replace it entirely.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_middleware']
        autoconfig.configure_settings(self.settings_dict)
        self.assertIn('my.middleware', self.settings_dict['MIDDLEWARE_CLASSES'])
        self.assertIn('django.middleware.common.CommonMiddleware', self.settings_dict['MIDDLEWARE_CLASSES'])

    def test_no_autoconfig(self):
        '''
        An app with no autoconfig shouldn't break things.
        '''
        self.settings_dict['INSTALLED_APPS'] = [
            'django_autoconfig.tests.app_no_autoconfig',
            'django',
        ]
        autoconfig.configure_settings(self.settings_dict)

    def test_blank_autoconfig(self):
        '''
        An app with a blank autoconfig shouldn't break things.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_blank_autoconfig']
        autoconfig.configure_settings(self.settings_dict)

    def test_booleans(self):
        '''
        Things we can't merge just get replaced.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_boolean']
        autoconfig.configure_settings(self.settings_dict)
        self.assertEqual(self.settings_dict['DEBUG'], True)

    def test_inconsistency(self):
        '''
        Check for required inconsistencies.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_boolean', 'django_autoconfig.tests.app_boolean_inconsistent']
        with self.assertRaises(ImproperlyConfigured):
            autoconfig.configure_settings(self.settings_dict)

    def test_relationship(self):
        '''
        Test putting things somewhere other than at the end of the list.
        '''
        self.settings_dict['INSTALLED_APPS'] = [
            'django_autoconfig.tests.app1',
            'django_autoconfig.tests.app2',
            'django_autoconfig.tests.app_relationship',
        ]
        autoconfig.configure_settings(self.settings_dict)
        self.assertEqual(
            self.settings_dict['INSTALLED_APPS'],
            [
                'django_autoconfig.tests.app_relationship',
                'django_autoconfig.tests.app1',
                'django_autoconfig.tests.app3',
                'django_autoconfig.tests.app2',
            ],
        )

    def test_default_setting(self):
        '''
        A setting in the DEFAULTS section should be used like merging.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_default_settings']
        autoconfig.configure_settings(self.settings_dict)
        self.assertEqual(self.settings_dict['DEFAULT_SETTING'], [1, 2, 3])

    def test_default_existing_setting(self):
        '''
        A setting in the DEFAULTS section should only be used if
        it doesn't already exist.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_default_settings']
        self.settings_dict['DEFAULT_SETTING'] = [4, 5, 6]
        autoconfig.configure_settings(self.settings_dict)
        self.assertEqual(self.settings_dict['DEFAULT_SETTING'], [4, 5, 6])

    def test_importerror_from_no_parent(self):
        '''
        An import error due to the parent module not existing should be raised.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['i.do.not.exist']
        with self.assertRaises(ImportError):
            autoconfig.configure_settings(self.settings_dict)

    def test_importerror_from_import_error(self):
        '''
        An import error due to the module itself generating an import error should be raised.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_broken_autoconfig']
        with self.assertRaises(ImportError) as exception_manager:
            autoconfig.configure_settings(self.settings_dict)
        self.assertIn('flibble', str(exception_manager.exception))

    def test_disabled_autoconfig(self):
        '''
        Test the ability to disable autoconfig for some apps.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django_autoconfig.tests.app_list']
        self.settings_dict['AUTOCONFIG_DISABLED_APPS'] = ['django_autoconfig.tests.app_list']
        autoconfig.configure_settings(self.settings_dict)
        self.assertEqual(self.settings_dict['LIST_SETTING'], [1, 2])

    def test_contrib_autoconfig(self):
        '''
        Test autoconfig provided by us instead of the app.
        '''
        self.settings_dict['INSTALLED_APPS'] = ['django.contrib.auth']
        autoconfig.configure_settings(self.settings_dict)
        self.assertIn('django.contrib.sessions', self.settings_dict['INSTALLED_APPS'])

    def test_logging_premature_imports(self):
        '''
        Test that logging doesn't cause premature imports.
        '''

        import logging, StringIO
        output = StringIO.StringIO()
        stream_handler = logging.StreamHandler(output)
        logger = logging.getLogger('django_autoconfig.autoconfig')
        logger.addHandler(stream_handler)
        logger.setLevel(logging.DEBUG)

        self.triggered = False
        class Trap(object):
            def __init__(self, sentinel):
                self.sentinel = sentinel
            def __str__(self):
                self.sentinel.triggered = True
                return 'triggered!'

        autoconfig.merge_dictionaries({}, {'LOGIN_URL': Trap(self)})
        self.assertFalse(self.triggered)

        logger.removeHandler(stream_handler)


class ConfigureUrlsTestCase(test.TestCase):
    '''Test the autoconfiguration of the urlconf.'''
    urls = 'django_autoconfig.autourlconf'

    def create_urlconf(self, apps):
        '''Create a urlconf from a list of apps.'''
        self.urlpatterns = autoconfig.configure_urls(apps)
        django.core.urlresolvers._resolver_cache = {}

    def test_urls(self):
        '''Test a simple url autoconfiguration.'''
        self.create_urlconf(['django_autoconfig.tests.app_urls'])
        resolve('/django-autoconfig.tests.app-urls/index/', urlconf=self)

    def test_blank_urls(self):
        '''Test a url autoconfiguration with an app with a blank urls.py.'''
        self.create_urlconf(['django_autoconfig.tests.app_urls', 'django_autoconfig.tests.app_blank_urls'])
        with self.assertRaises(django.core.urlresolvers.Resolver404):
            resolve('/django-autoconfig.tests.app-blank-urls/index/', urlconf=self)

    def test_missing_app_urls(self):
        '''Test a url autoconfiguration with an app without urls.'''
        self.create_urlconf(['django_autoconfig.tests.app_urls', 'django_autoconfig.tests.app1'])
        with self.assertRaises(django.core.urlresolvers.Resolver404):
            resolve('/django-autoconfig.tests.app1/index/', urlconf=self)



'''Automatic configuration for Django project.'''

import collections
import copy
from django.core.exceptions import ImproperlyConfigured
from django.conf import global_settings
from django.conf.urls import include, patterns, url
from django.utils.module_loading import module_has_submodule
import importlib
import operator

import logging
LOGGER = logging.getLogger(__name__)

MAX_ITERATIONS = 1000

SETTINGS = {
    'AUTOCONFIG_DISABLED_APPS': [
        'django.contrib.admin',
        'django.contrib.auth',
    ],
}

class OrderingRelationship(object):
    '''
    This class defines a relationship between an element in a setting
    that's a list and one or more other entries.

    It's intended to be used in an autoconfig.py file like so::

        RELATIONSHIPS = [
            OrderingRelationship(
                'INSTALLED_APPS',
                'my.app',
                before = [
                    'django.contrib.admin',
                ],
                after = [
                ],
            )
        ]
    '''

    def __init__(
        self,
        setting_name,
        setting_value,
        before=None,
        after=None,
        add_missing=True
    ):
        self.setting_name = setting_name
        self.setting_value = setting_value
        self.before = before or []
        self.after = after or []
        self.add_missing = add_missing

    def apply_changes(self, settings):
        changes = 0

        if self.add_missing:
            for item in [self.setting_value] + self.before + self.after:
                if item not in settings[self.setting_name]:
                    settings[self.setting_name] = list(settings[self.setting_name]) + [item]
                    LOGGER.debug("Added %r to %r.", item, self.setting_name)
                    changes += 1
        elif self.setting_value not in settings[self.setting_name]:
            return changes

        for test, related_items, list_name in (
            (operator.gt, self.before, 'before'),
            (operator.lt, self.after, 'after'),
        ):
            current_value = settings[self.setting_name]

            for item in related_items:
                if item not in current_value:
                    continue
                if test(
                    current_value.index(self.setting_value),
                    current_value.index(item),
                ):
                    if isinstance(current_value, tuple):
                        current_value = list(current_value)
                    location = current_value.index(item)
                    current_value.remove(self.setting_value)
                    current_value.insert(location, self.setting_value)
                    settings[self.setting_name] = current_value
                    LOGGER.debug("Moved %r %r %r.", self.setting_value, list_name, item)
                    changes += 1

        return changes

def merge_dictionaries(current, new, only_defaults=False):
    '''
    Merge two settings dictionaries, recording how many changes were needed.

    '''
    changes = 0
    for key, value in new.items():
        if key not in current:
            if hasattr(global_settings, key):
                current[key] = getattr(global_settings, key)
                LOGGER.debug("Set %r to global default %r.", key, current[key])
            else:
                current[key] = copy.copy(value)
                LOGGER.debug("Set %r to %r.", key, current[key])
                changes += 1
                continue
        elif only_defaults:
            continue
        current_value = current[key]
        if hasattr(current_value, 'items'):
            changes += merge_dictionaries(current_value, value)
        elif isinstance(current_value, (list, tuple)):
            for element in value:
                if element not in current_value:
                    current[key] = list(current_value) + [element]
                    LOGGER.debug("Added %r to %r.", element, key)
                    changes += 1
        else:
            # If we don't know what to do with it, replace it.
            if current_value != value:
                current[key] = value
                LOGGER.debug("Set %r to %r.", key, current[key])
                changes += 1
    return changes

def configure_settings(settings):
    '''
    Given a settings object, run automatic configuration of all
    the apps in INSTALLED_APPS.
    '''
    changes = 1
    iterations = 0

    while changes:
        changes = 0
        for app_name in list(settings['INSTALLED_APPS']) + ['django_autoconfig']:
            if app_name not in settings.get('AUTOCONFIG_DISABLED_APPS', ()):
                app_module = importlib.import_module(app_name)
            else:
                app_module = None
            import django_autoconfig.contrib
            if app_module and module_has_submodule(app_module, 'autoconfig'):
                module = importlib.import_module("%s.autoconfig" % (app_name,))
            elif app_name in django_autoconfig.contrib.CONTRIB_CONFIGS:
                module = django_autoconfig.contrib.CONTRIB_CONFIGS[app_name]
            else:
                continue
            changes += merge_dictionaries(
                settings,
                getattr(module, 'SETTINGS', {}),
            )
            changes += merge_dictionaries(
                settings,
                getattr(module, 'DEFAULT_SETTINGS', {}),
                only_defaults=True,
            )
            for relationship in getattr(module, 'RELATIONSHIPS', []):
                changes += relationship.apply_changes(settings)

        if iterations >= MAX_ITERATIONS:
            raise ImproperlyConfigured(
                'Autoconfiguration could not reach a consistent state'
            )
        iterations += 1
    LOGGER.debug("Autoconfiguration took %d iterations.", iterations)

def configure_urls(apps):
    '''
    Configure urls from a list of apps.
    '''
    urlpatterns = patterns('')

    for app_name in apps:
        try:
            module = importlib.import_module("%s.urls" % app_name)
            if not hasattr(module, 'urlpatterns'):
                # Resolver will break if the urls.py file is completely blank.
                continue
            urlpatterns += patterns(
                '',
                url(
                    r'^%s/' % app_name.replace("_","-"),
                    include("%s.urls" % app_name),
                ),
            )
        except ImportError:
            pass
    return urlpatterns



'''This module can be set as a ROOT_URLCONF (or included into one).'''

from django.conf import settings

from .app_settings import AUTOCONFIG_EXTRA_URLS
from .autoconfig import configure_urls

urlpatterns = configure_urls(list(settings.INSTALLED_APPS) + list(AUTOCONFIG_EXTRA_URLS)) # pylint: disable=C0103



'''Autoconfig data for applications that don't support this protocol.'''

from collections import namedtuple

from django_autoconfig.autoconfig import OrderingRelationship

Autoconfig = namedtuple('Autoconfig', ('SETTINGS', 'DEFAULT_SETTINGS', 'RELATIONSHIPS'))

CONTRIB_CONFIGS = {
    'django.contrib.auth': Autoconfig(
        SETTINGS = {
            'INSTALLED_APPS': [
                'django.contrib.contenttypes',
                'django.contrib.sessions',
            ],
            'MIDDLEWARE_CLASSES': [
                'django.contrib.auth.middleware.AuthenticationMiddleware',
            ],
        },
        DEFAULT_SETTINGS = {},
        RELATIONSHIPS = [
            OrderingRelationship(
                'MIDDLEWARE_CLASSES',
                'django.contrib.auth.middleware.AuthenticationMiddleware',
                after = [
                    'django.middleware.common.CommonMiddleware',
                    'django.contrib.sessions.middleware.SessionMiddleware',
                    'django.middleware.csrf.CsrfViewMiddleware',
                ],
                add_missing = False,
            )
        ],
    ),
    'django.contrib.messages': Autoconfig(
        SETTINGS = {
            'INSTALLED_APPS': [
                'django.contrib.sessions',
            ],
            'MIDDLEWARE_CLASSES': [
                'django.contrib.messages.middleware.MessageMiddleware',
            ],
            'TEMPLATE_CONTEXT_PROCESSORS': [
                'django.contrib.messages.context_processors.messages',
            ],
        },
        DEFAULT_SETTINGS = {},
        RELATIONSHIPS = [
            OrderingRelationship(
                'MIDDLEWARE_CLASSES',
                'django.contrib.messages.middleware.MessageMiddleware',
                after = [
                    'django.contrib.sessions.middleware.SessionMiddleware',
                    'django.contrib.auth.middleware.AuthenticationMiddleware',
                ],
                add_missing = False,
            )
        ],
    ),
    'django.contrib.sessions': Autoconfig(
        SETTINGS = {
            'MIDDLEWARE_CLASSES': [
                'django.contrib.sessions.middleware.SessionMiddleware',
            ],
        },
        DEFAULT_SETTINGS = {},
        RELATIONSHIPS = [
            OrderingRelationship(
                'MIDDLEWARE_CLASSES',
                'django.contrib.sessions.middleware.SessionMiddleware',
                after = [
                    'django.middleware.cache.UpdateCacheMiddleware',
                ],
                before = [
                    'django.middleware.common.CommonMiddleware',
                ],
                add_missing = False,
            )
        ],
    ),
    'django.contrib.admin': Autoconfig(
        SETTINGS = {
            'INSTALLED_APPS': [
                'django.contrib.auth',
                'django.contrib.contenttypes',
                'django.contrib.messages',
            ],
            'AUTOCONFIG_EXTRA_URLS': [
                'django_autoconfig.contrib.admin',
            ],
        },
        DEFAULT_SETTINGS = {},
        RELATIONSHIPS = [
            OrderingRelationship(
                'INSTALLED_APPS',
                'django.contrib.admin',
                after = [
                    'django.contrib.contenttypes',
                ],
                add_missing = False,
            )
        ],
    ),
}
